import numpy as np
import pandas as pd
import pytest

d1 = np.random.randint(0, 100, size=(4, 3))
d2 = np.random.randint(0, 100, size=(2, 3))
df1 = pd.DataFrame(d1, columns=list('ABC'))
df2 = pd.DataFrame(d2, columns=list('ABC'), index=[100, 200])
c = [pd.concat((df1, df2), axis=0), pd.concat((df1, df2), axis=1)]



### Begin testsing
import dovpanda
@pytest.mark.parametrize('axis', [0, 1])
def test_wrong_concat_axis(axis):
    expected = c[axis]
    result = pd.concat((df1, df2), axis=axis)
    pd.testing.assert_frame_equal(result, expected)


__author__ = """Dean Langsam"""
__version__ = '0.0.3.beta'

import sys

from dovpanda.core import ledger

if 'pandas' in sys.modules.keys():
    ledger.register_hints()
else:
    ledger.tell('Pandas not imported')


def set_output(tell_method):
    ledger.set_output(tell_method)



from dovpanda import base
from dovpanda.base import Ledger

ledger = Ledger()


@ledger.add_hint('DataFrame.__init__')
def init_for_checks(*args, **kwargs):
    ledger.tell('you have construted a df')


@ledger.add_hint('DataFrame.__init__')
def init_another(*args, **kwargs):
    ledger.tell('another pre hook for init')


@ledger.add_hint('DataFrame.iterrows')
def iterrows_is_bad(*args, **kwargs):
    ledger.tell("iterrows is not recommended, and in the majority of cases will have better alternatives")


@ledger.add_hint('DataFrame.groupby')
def time_grouping(*args, **kwargs):
    try:
        by = args[1]
    except IndexError:
        by = kwargs.get('by')
    base.listify(by)
    if 'hour' in by:
        ledger.tell('Seems like you are grouping by time, consider using resample')


@ledger.add_hint('concat', hook_type='post')
def duplicate_index_after_concat(res, *args, **kwargs):
    if res.index.nunique() != len(res.index):
        ledger.tell('After concatenation you have duplicated indexes values - pay attention')
    if res.columns.nunique() != len(res.columns):
        ledger.tell('After concatenation you have duplicated column names - pay attention')


@ledger.add_hint('concat')
def concat_single_column(*args, **kwargs):
    objs = base.get_arg(args, kwargs, 0, 'objs')
    axis = base.get_arg(args, kwargs, 1, 'axis')
    cols = {df.shape[1] for df in objs}
    if axis == 1 and 1 in cols:
        ledger.tell(
            'One of the dataframes you are concatenating is with a single column, '
            'consider using `df.assign()` or `df.insert()`')


@ledger.add_hint('concat')
def wrong_concat_axis(*args, **kwargs):
    objs = base.get_arg(args, kwargs, 0, 'objs')
    axis = base.get_arg(args, kwargs, 1, 'axis')
    rows = {df.shape[0] for df in objs}
    cols = {df.shape[1] for df in objs}
    col_names = set.union(*[set(df.columns) for df in objs])
    same_cols = (len(cols) == 1) and (len(col_names) == list(cols)[0])
    same_rows = (len(rows) == 1)
    axis_translation = {0: 'vertically', 1: 'horizontally'}
    if same_cols and not same_rows:
        if axis == 1:
            ledger.tell("All dataframes have the same columns, which hints for concat on axis 0."
                        "You specified <code>axis=1</code> which may result in an unwanted behaviour")
    elif same_rows and not same_cols:
        if axis == 0:
            ledger.tell("All dataframes have same number of rows, which hints for concat on axis 1."
                        "You specified <code>axis=0</code> which may result in an unwanted behaviour")

    elif same_rows and same_rows:
        ledger.tell("All dataframes have the same columns and same number of rows. "
                    f"Pay attention, your axis is {axis} which concatenates {axis_translation[axis]}")


@ledger.add_hint('DataFrame.__eq__')
def df_check_equality(*args):
    if isinstance(args[0], type(args[1])):
        ledger.tell(f'Calling df1 == df2 compares the objects element-wise. '
                    'If you need a boolean condition, try df1.equals(df2)')


@ledger.add_hint('Series.__eq__')
def series_check_equality(*args):
    if isinstance(args[0], type(args[1])):
        ledger.tell(f'Calling series1 == series2 compares the objects element-wise. '
                    'If you need a boolean condition, try series1.equals(series2)')


@ledger.add_hint('read_csv', 'post')
def csv_index(res, *args, **kwargs):
    filename = base.get_arg(args, kwargs, 0, 'filepath_or_buffer')
    if type(filename) is str:
        filename = "'" + filename + "'"
    else:
        filename = 'file'
    if 'Unnamed: 0' in res.columns:
        if (len(args) < 5) and ('index_col' not in kwargs.keys()):
            ledger.tell('Your left most column is unnamed. This suggets it might be the index column, try: '
                        f'<code>pd.read_csv({filename}, index_col=0)</code>')



import functools
import pathlib
import re
import sys
import traceback
from collections import defaultdict

import pandas

PANDAS_DIR = str(pathlib.Path(pandas.__file__).parent.absolute())
try:  # If user runs from notebook they will have this
    from IPython.display import display
except (ModuleNotFoundError, ImportError):
    pass


class Hint:
    def __init__(self, original, hook_type, replacement):
        accepted_hooks = ['pre', 'post']
        assert hook_type in accepted_hooks, f'hook_type must be one of {accepted_hooks}'

        self.original = original
        self.hook_type = hook_type
        self.replacement = replacement


class _Teller:
    def __init__(self):
        self.message = None
        self.set_output('display')
        self.line = None
        self.lineno = None
        self.verbose = True

    def __repr__(self):
        trace = self.if_verbose(f' (Line {self.lineno})')
        return self._strip_html(f'===== {self.message} ====={trace}')

    def __str__(self):
        trace = self.if_verbose(f' (Line {self.lineno})')
        return f'{self._strip_html(self.message)}{trace}'

    def __call__(self, s):
        self.tell(s)

    def _repr_html_(self):
        return self.nice_output()

    def nice_output(self):
        trace = f'<div style="font-size:0.7em;">Line {self.lineno}: <code>{self.line}</code> </div>'
        trace = self.if_verbose(trace)
        html = f'''
        <div class="alert alert-info" role="alert">
          {self.message}
          <button type="button" class="close" data-dismiss="alert" aria-label="Close">
            <span aria-hidden="true">&times;</span>
          </button>
          {trace}

        </div>
        '''
        return html

    @staticmethod
    def _strip_html(s):
        return re.sub('<[^<]+?>', '', s)

    @staticmethod
    def _no_output(*args, **kwargs):
        return

    def if_verbose(self, s):
        if self.verbose:
            return s
        else:
            return ''

    def set_output(self, output_method):
        accepted_methods = ['print', 'display', 'debug', 'info', 'warning', 'off']
        if type(output_method) is str:
            assert output_method in accepted_methods, f'output_format must be one of {accepted_methods} or a callable'
        if output_method == 'print':
            self.output = print
        elif output_method in ['debug', 'info', 'warning']:
            try:
                from loguru import logger
            except ModuleNotFoundError:
                import logging
                logger = logging.getLogger('dovpanda')
            self.output = getattr(logger, output_method)
        elif output_method == 'display':
            try:
                self.output = display
            except (ModuleNotFoundError, ImportError):
                self.set_output('print')
        elif output_method == 'off':
            self.output = self._no_output
        else:
            self.output = output_method

    def tell(self, message):
        self.message = message
        self.output(self)


class Ledger:
    def __init__(self):
        self.hints = defaultdict(list)
        self.teller = _Teller()
        self.verbose = True
        self.caller_filename = None

    def replace(self, original, func_hooks: tuple):
        g = rgetattr(sys.modules['pandas'], original)
        rsetattr(sys.modules['pandas'], original, self.attach_hooks(g, func_hooks))

    def add_hint(self, original, hook_type='pre'):

        def replaces_decorator(replacement):
            hint = Hint(original=original, hook_type=hook_type, replacement=replacement)
            self.hints[original].append(hint)

        return replaces_decorator

    def register_hints(self):
        for original, func_hooks in self.hints.items():
            self.replace(original, func_hooks)

    def attach_hooks(self, f, func_hooks):
        pres = [hook.replacement for hook in func_hooks if hook.hook_type == 'pre']
        posts = [hook.replacement for hook in func_hooks if hook.hook_type == 'post']

        @functools.wraps(f)
        def run(*args, **kwargs):
            caller = traceback.extract_stack()[-2]
            self._set_caller_details(caller)
            if self.inner_pandas:
                ret = f(*args, **kwargs)
            else:
                for pre in pres:
                    pre(*args, **kwargs)
                ret = f(*args, **kwargs)
                for post in posts:
                    post(ret, *args, **kwargs)
            return ret

        return run

    def _set_caller_details(self, caller):
        self.caller_filename = caller.filename
        self.inner_pandas = self.caller_filename.startswith(PANDAS_DIR)
        if self.caller_filename.startswith(PANDAS_DIR):
            return
        self.teller.line = caller.line
        self.teller.lineno = caller.lineno

    # Output
    def tell(self, *args, **kwargs):
        self.teller(*args, *kwargs)

    def set_output(self, output):
        self.teller.set_output(output)

    def set_verbose(self, verbose=True):
        self.teller.verbose = verbose


def get_arg(args, kwargs, which_arg, which_kwarg):
    try:
        ret = args[which_arg]
    except IndexError:
        ret = kwargs.get(which_kwarg)
    return ret


def rgetattr(obj, attr):
    attributes = attr.strip('.').split('.')
    for att in attributes:
        try:
            obj = getattr(obj, att)
        except AttributeError as e:
            raise e
    return obj


def rsetattr(obj, attr, value):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, value)


def only_print(s, *args, **kwargs):
    print(s)
    return lambda x: (args, kwargs)


def listify(val):
    if type(val) is str:
        return [val]
    return val


def setify(val):
    return set(listify(val))



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.md').read()
doclink = """
Documentation
-------------

The full documentation is at http://dovpanda.rtfd.org."""
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name="dovpanda",
    version='0.0.3.beta',
    description='Directions overlay for working with pandas in an analysis environment',
    author="Dean Langsam",
    author_email="deanla@gmail.com",
    long_description=readme + '\n\n' + doclink + '\n\n' + history,
    long_description_content_type="text/markdown",
    url="https://github.com/dovpanda-dev/dovpanda",
    packages=['dovpanda'],
    install_requires=['pandas'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

    ],
    python_requires='>=3.6',
)
